/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * Copyright (C) 2011-2012 EasyRPG Project
 *
 * License: https://github.com/EasyRPG/Editor/blob/master/COPYING GPL
 *
 * Authors:
 * - Aitor García (Falc) <aitor.falc@gmail.com>
 * - Mariano Suligoy (MarianoGNU) <marianognu.easyrpg@gmail.com>
 * - Francisco de la Peña (fdelapena) <fran@fran.cr>
 * - Sebastian Reichel (sre) <sre@ring0.de>
 */

/**
 * XmlServer is a tool that uses GLib.MarkupParser to parse XML files and convert their
 * string-formatted content to XmlNode instances.
 */
public class XmlParser {
	// Since Vala 0.14 parser started to crash. The solution is declare it as const
	const GLib.MarkupParser parser = {
		opening_tag_callback,
		closing_tag_callback,
		tag_content_callback,
		passthrough_callback,
		error_callback
	};

	/*
	 * Properties
	 */
	private GLib.MarkupParseContext context;
	private int nesting_level = 0;

	/**
	 * A reference to the root XmlNode of the last parsed file.
	 */
	private XmlNode root;

	/*
	 * An auxiliar reference used to point at XmlNodes when parsing.
	 */
	private XmlNode current_ref;

	/**
	 * Instantiates the parse context.
	 */
	public XmlParser () {
		this.context = new GLib.MarkupParseContext (
			parser, // MarkupParser instance
			0,		// MarkupParseFlags
			this,   // User data
			null	// User data destroy notifier
		);
	}

	/**
	 * Converts the XML file to a hierarchically organized list of XmlNodes.
	 *
	 * @param path The path to the XML file.
	 */
	public void parse_file (string path) throws Error {
		this.root = null;

		//GLib.File file = GLib.File.new_for_path (path);

		string file_content;

		if (GLib.FileUtils.get_contents (path, out file_content)) {
			file_content = Utils.clean_file_content(file_content);

			this.context.parse (file_content, -1);
		}
	}

	/*
	 * This method is called each time the parser finds an opening tag.
	 *
	 * The parser defines some parameters containing the information related to the tag.
	 */
	private void opening_tag_callback (GLib.MarkupParseContext ctx, string tag_name,
	                                   string[] attr_names, string[] attr_values)
                                      throws GLib.MarkupError {
		/*
		 * Root Node
		 *
		 * Only one root level block is parsed. Any other root level block
		 * found in the file is automatically ignored.
		 */
		if(this.nesting_level == 0) {
			// If a root block was already parsed, ignore the rest of the file
			if(this.root != null && this.root.name != "") {
				ctx.end_parse ();
				return;
			}

			this.root = new XmlNode ();
			this.root.name = tag_name;
			this.root.parent = null;
			this.root.prev = null;
			this.root.next = null;
			this.current_ref = this.root;
		}
		/*
		 * Non-root node
		 */
		else {
			XmlNode node = new XmlNode ();
			node.name = tag_name;
			for(int i = 0; i < attr_names.length ; i++)
				node.attributes[attr_names[i]] = attr_values[i];
			node.parent = this.current_ref;

			this.current_ref.add_child (node);
			this.current_ref = node;
		}

		this.nesting_level++;
	}

	/*
	 * This method is called each time the parser finds a closing tag.
	 *
	 * The parser defines some parameters containing the information related to the tag.
	 */
	private void closing_tag_callback (GLib.MarkupParseContext ctx, string tag_name)
                                      throws GLib.MarkupError {
		this.current_ref = this.current_ref.parent;
		this.nesting_level--;
	}

	/*
	 * This method is called each time the parser finds content inside tags.
	 */
	private void tag_content_callback (GLib.MarkupParseContext ctx, string content, size_t content_len)
                                      throws GLib.MarkupError {
		this.current_ref.content = content;
	}

	/*
	 * This method is called each time the parser finds comments, processing instructions
	 * and doctype declarations.
	 */
	private void passthrough_callback (GLib.MarkupParseContext ctx, string passthrough_text, size_t text_len)
                                      throws GLib.MarkupError {
		// Not used yet
	}

	/*
	 * This method is called when the parser finds an error.
	 */
	private void error_callback (GLib.MarkupParseContext ctx, GLib.Error error) {
		// Not used yet
	}

	/**
	 * Returns the root of the tree generated from the XML data.
	 */
	public XmlNode get_root () {
		return this.root;
	}
}
